from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np
from decimal import Decimal, ROUND_HALF_UP, ROUND_HALF_EVEN


import nngen.util as util


def avg_pool(value, ksize, stride, padding='SAME',
             dtype=None, sum_dtype=None, name=None, par=1,
             force_div=False,
             value_ram_size=None, out_ram_size=None,
             value_dtype=None):

    ksize_row = ksize[1]
    ksize_col = ksize[2]
    stride_row = stride[2]
    stride_col = stride[2]

    if isinstance(padding, str) and padding == 'SAME':
        pad_col, pad_col_left, pad_col_right = util.pad_size_split(
            value.shape[2], ksize_col, stride_col)
        pad_row, pad_row_top, pad_row_bottom = util.pad_size_split(
            value.shape[1], ksize_row, stride_row)

        out_shape = (value.shape[0],
                     util.pix_size(value.shape[1], ksize_row, stride_row, padding),
                     util.pix_size(value.shape[2], ksize_col, stride_col, padding),
                     value.shape[3])

    elif isinstance(padding, str) and padding == 'VALID':
        pad_col, pad_col_left, pad_col_right = 0, 0, 0
        pad_row, pad_row_top, pad_row_bottom = 0, 0, 0

        out_shape = (value.shape[0],
                     util.pix_size(value.shape[1], ksize_row, stride_row, 'VALID'),
                     util.pix_size(value.shape[2], ksize_col, stride_col, 'VALID'),
                     value.shape[3])

    elif isinstance(padding, int):
        pad_col, pad_col_left, pad_col_right = padding * 2, padding, padding
        pad_row, pad_row_top, pad_row_bottom = padding * 2, padding, padding

        out_shape = (value.shape[0],
                     util.pix_size(value.shape[1] + padding * 2,
                                   ksize_row, stride_row, 'VALID'),
                     util.pix_size(value.shape[2] + padding * 2,
                                   ksize_col, stride_col, 'VALID'),
                     value.shape[3])

    elif isinstance(padding, (tuple, list)):
        pad_col, pad_col_left, pad_col_right = padding[2] + padding[3], padding[2], padding[3]
        pad_row, pad_row_top, pad_row_bottom = padding[0] + padding[1], padding[0], padding[1]

        out_shape = (value.shape[0],
                     util.pix_size(value.shape[1] + padding[0] + padding[1],
                                   ksize_row, stride_row, 'VALID'),
                     util.pix_size(value.shape[2] + padding[2] + padding[3],
                                   ksize_col, stride_col, 'VALID'),
                     value.shape[3])

    else:
        raise ValueError("padding options must be 'SAME', 'VALID', int, tuple, or list.")

    out = np.zeros(out_shape, dtype=np.int64)

    value = np.pad(value, [(0, 0),
                           (pad_row_top, pad_row_bottom),
                           (pad_col_left, pad_col_right),
                           (0, 0)], 'constant')

    value_point = 0 if value_dtype is None else value_dtype.point
    out_point = value_point if dtype is None else dtype.point
    div_shift = out_point - value_point

    div_op = (lambda x: x << div_shift if div_shift >= 0 else
              lambda x: x >> -div_shift)

    num_vars = ksize_col * ksize_row

    def divider(vx):
        return np.array(
            list(map(lambda x: Decimal(str(x / num_vars)).quantize(Decimal('0'),
                                                                   rounding=ROUND_HALF_UP), vx))).astype(np.int64)

    for bat in range(value.shape[0]):

        oy = 0
        for py in range(-pad_row_top, value.shape[1] + pad_row_bottom, stride_row):

            ox = 0
            for px in range(-pad_col_left, value.shape[2] + pad_col_right, stride_col):

                ys = py + pad_row_top
                ye = ys + ksize_row
                xs = px + pad_col_left
                xe = xs + ksize_col
                a = value[bat, ys: ye, xs: xe]
                a = np.add.reduce(a, axis=0)
                sum = np.add.reduce(a, axis=0)
                div = divider(sum)

                out[bat][oy][ox][:] = div_op(div)

                ox += 1
                if ox >= out.shape[2]:
                    break

            oy += 1
            if oy >= out.shape[1]:
                break

    return out


def max_pool(value, ksize, stride, padding='SAME',
             dtype=None, name=None, par=1,
             value_ram_size=None, out_ram_size=None,
             value_dtype=None):

    ksize_row = ksize[1]
    ksize_col = ksize[2]
    stride_row = stride[2]
    stride_col = stride[2]

    if isinstance(padding, str) and padding == 'SAME':
        pad_col, pad_col_left, pad_col_right = util.pad_size_split(
            value.shape[2], ksize_col, stride_col)
        pad_row, pad_row_top, pad_row_bottom = util.pad_size_split(
            value.shape[1], ksize_row, stride_row)

        out_shape = (value.shape[0],
                     util.pix_size(value.shape[1], ksize_row, stride_row, padding),
                     util.pix_size(value.shape[2], ksize_col, stride_col, padding),
                     value.shape[3])

    elif isinstance(padding, str) and padding == 'VALID':
        pad_col, pad_col_left, pad_col_right = 0, 0, 0
        pad_row, pad_row_top, pad_row_bottom = 0, 0, 0

        out_shape = (value.shape[0],
                     util.pix_size(value.shape[1], ksize_row, stride_row, 'VALID'),
                     util.pix_size(value.shape[2], ksize_col, stride_col, 'VALID'),
                     value.shape[3])

    elif isinstance(padding, int):
        pad_col, pad_col_left, pad_col_right = padding * 2, padding, padding
        pad_row, pad_row_top, pad_row_bottom = padding * 2, padding, padding

        out_shape = (value.shape[0],
                     util.pix_size(value.shape[1] + padding * 2,
                                   ksize_row, stride_row, 'VALID'),
                     util.pix_size(value.shape[2] + padding * 2,
                                   ksize_col, stride_col, 'VALID'),
                     value.shape[3])

    elif isinstance(padding, (tuple, list)):
        pad_col, pad_col_left, pad_col_right = padding[2] + padding[3], padding[2], padding[3]
        pad_row, pad_row_top, pad_row_bottom = padding[0] + padding[1], padding[0], padding[1]

        out_shape = (value.shape[0],
                     util.pix_size(value.shape[1] + padding[0] + padding[1],
                                   ksize_row, stride_row, 'VALID'),
                     util.pix_size(value.shape[2] + padding[2] + padding[3],
                                   ksize_col, stride_col, 'VALID'),
                     value.shape[3])

    else:
        raise ValueError("padding options must be 'SAME', 'VALID', int, tuple, or list.")

    out = np.zeros(out_shape, dtype=np.int64)

    if value_dtype is not None:
        pad_value_shift = value_dtype.width
    elif dtype is not None:
        pad_value_shift = dtype.width
    else:
        pad_value_shift = 32
    pad_value = (-1) * (1 << (pad_value_shift - 1))

    value = np.pad(value, [(0, 0),
                           (pad_row_top, pad_row_bottom),
                           (pad_col_left, pad_col_right),
                           (0, 0)], 'constant',
                   constant_values=pad_value)

    value_point = 0 if value_dtype is None else value_dtype.point
    out_point = value_point if dtype is None else dtype.point
    max_shift = out_point - value_point

    max_op = ((lambda x: x << max_shift) if max_shift >= 0 else
              (lambda x: x >> -max_shift))

    for bat in range(value.shape[0]):

        oy = 0
        for py in range(-pad_row_top, value.shape[1] + pad_row_bottom, stride_row):

            ox = 0
            for px in range(-pad_col_left, value.shape[2] + pad_col_right, stride_col):

                ys = py + pad_row_top
                ye = ys + ksize_row
                xs = px + pad_col_left
                xe = xs + ksize_col
                a = value[bat, ys: ye, xs: xe]
                a = np.max(a, axis=0)
                max_val = np.max(a, axis=0)

                out[bat][oy][ox][:] = max_val

                ox += 1
                if ox >= out.shape[2]:
                    break

            oy += 1
            if oy >= out.shape[1]:
                break

    return out



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np


def exp(visitor, node):
    features = node.args[0]

    visitor.visit(features)

    addr_scale = 1 / node._get_expected_scale_factor()

    out_width = node.dtype.width
    out_point = node.dtype.point
    out_signed = node.dtype.signed
    if out_signed:
        out_scale = ((2 ** (out_width - 1)) * node.range_rate /
                     np.exp(2 ** (node.lut_addrwidth - 1) * addr_scale + node.lut_bias))
    else:
        out_scale = ((2 ** out_width) * node.range_rate /
                     np.exp(2 ** (node.lut_addrwidth - 1) * addr_scale + node.lut_bias))

    if out_point == 0:
        out_scale = out_scale
    elif out_point > 0:
        out_scale = out_scale / (2 ** out_point)
    else:
        out_scale = out_scale * (2 ** (-1 * out_point))

    node.scale_factor = out_scale



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division


def argmax(visitor, node):

    input_tensor = node.args[0]
    visitor.visit(input_tensor)

    node.scale_factor = 1.0


def argmin(visitor, node):

    input_tensor = node.args[0]
    visitor.visit(input_tensor)

    node.scale_factor = 1.0



from __future__ import absolute_import
from __future__ import print_function

import os
import sys

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.abspath(__file__)))))

import nngen as ng
import veriloggen

import matrix_conv2d_conv2d


act_shape = (1, 7, 7, 15)
weight1_shape = (7, 1, 1, 15)
bias1_shape = None
scale1_shape = None
weight2_shape = (9, 1, 1, 7)
bias2_shape = None
scale2_shape = None
act_dtype = ng.int16
weight1_dtype = ng.int16
bias1_dtype = ng.int16
scale1_dtype = ng.int16
weight2_dtype = ng.int16
bias2_dtype = ng.int16
scale2_dtype = ng.int16
tmp_dtype = ng.int16
out_dtype = ng.int16
stride1 = (1, 1, 1, 1)
stride2 = (1, 1, 1, 1)
rshift_mul1 = None
rshift_sum1 = None
rshift_out1 = None
rshift_mul2 = None
rshift_sum2 = None
rshift_out2 = None
act_func1 = None
act_func2 = None
par_ich1 = 2
par_och1 = 2
par_col1 = 1
par_row1 = 1
concur_och1 = 8
stationary1 = 'filter'
input_ram_size1 = None
filter_ram_size1 = None
bias_ram_size1 = None
scale_ram_size1 = None
out_ram_size1 = None
par_ich2 = 2
par_och2 = 2
par_col2 = 1
par_row2 = 1
concur_och2 = 8
stationary2 = 'filter'
input_ram_size2 = None
filter_ram_size2 = None
bias_ram_size2 = None
scale_ram_size2 = None
out_ram_size2 = None
axi_datawidth = 32


def test(request, silent=True):
    veriloggen.reset()

    simtype = request.config.getoption('--sim')

    rslt = matrix_conv2d_conv2d.run(act_shape,
                                    weight1_shape, bias1_shape, scale1_shape,
                                    weight2_shape, bias2_shape, scale2_shape,
                                    act_dtype,
                                    weight1_dtype, bias1_dtype, scale1_dtype,
                                    weight2_dtype, bias2_dtype, scale2_dtype,
                                    tmp_dtype,
                                    out_dtype,
                                    stride1, stride2,
                                    rshift_mul1, rshift_sum1, rshift_out1,
                                    rshift_mul2, rshift_sum2, rshift_out2,
                                    act_func1, act_func2,
                                    par_ich1, par_och1, par_col1, par_row1,
                                    concur_och1, stationary1,
                                    par_ich2, par_och2, par_col2, par_row2,
                                    concur_och2, stationary2,
                                    input_ram_size1, filter_ram_size1,
                                    bias_ram_size1, scale_ram_size1,
                                    out_ram_size1,
                                    input_ram_size2, filter_ram_size2,
                                    bias_ram_size2, scale_ram_size2,
                                    out_ram_size2,
                                    axi_datawidth, silent,
                                    filename=None, simtype=simtype,
                                    outputfile=os.path.splitext(os.path.basename(__file__))[0] + '.out')

    verify_rslt = [line for line in rslt.splitlines() if line.startswith('# verify:')][0]
    assert(verify_rslt == '# verify: PASSED')


if __name__ == '__main__':
    rslt = matrix_conv2d_conv2d.run(act_shape,
                                    weight1_shape, bias1_shape, scale1_shape,
                                    weight2_shape, bias2_shape, scale2_shape,
                                    act_dtype,
                                    weight1_dtype, bias1_dtype, scale1_dtype,
                                    weight2_dtype, bias2_dtype, scale2_dtype,
                                    tmp_dtype,
                                    out_dtype,
                                    stride1, stride2,
                                    rshift_mul1, rshift_sum1, rshift_out1,
                                    rshift_mul2, rshift_sum2, rshift_out2,
                                    act_func1, act_func2,
                                    par_ich1, par_och1, par_col1, par_row1,
                                    concur_och1, stationary1,
                                    par_ich2, par_och2, par_col2, par_row2,
                                    concur_och2, stationary2,
                                    input_ram_size1, filter_ram_size1,
                                    bias_ram_size1, scale_ram_size1,
                                    out_ram_size1,
                                    input_ram_size2, filter_ram_size2,
                                    bias_ram_size2, scale_ram_size2,
                                    out_ram_size2,
                                    axi_datawidth, silent=False,
                                    filename='tmp.v',
                                    outputfile=os.path.splitext(os.path.basename(__file__))[0] + '.out')
    print(rslt)



from __future__ import absolute_import
from __future__ import print_function

import os
import sys

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.abspath(__file__)))))

import nngen as ng
import veriloggen

import onnx_matrix_conv2d


act_shape = (1, 7, 7, 3)
weight_shape = (9, 3, 3, 3)
act_dtype = ng.int32
weight_dtype = ng.int32
stride = 1
padding = 0
with_batchnorm = True
act_func = 'ReLU'
disable_fusion = False
par_ich = 1
par_och = 1
par_col = 1
par_row = 1
concur_och = None
stationary = 'filter'
chunk_size = 64
axi_datawidth = 32


def test(request, silent=True):
    veriloggen.reset()

    simtype = request.config.getoption('--sim')

    rslt = onnx_matrix_conv2d.run(act_shape, weight_shape,
                                  act_dtype, weight_dtype,
                                  stride, padding,
                                  with_batchnorm, act_func, disable_fusion,
                                  par_ich, par_och, par_col, par_row,
                                  concur_och, stationary,
                                  chunk_size,
                                  axi_datawidth, silent,
                                  filename=None, simtype=simtype,
                                  outputfile=os.path.splitext(os.path.basename(__file__))[0] + '.out')

    verify_rslt = [line for line in rslt.splitlines() if line.startswith('# verify:')][0]
    assert(verify_rslt == '# verify: PASSED')


if __name__ == '__main__':
    rslt = onnx_matrix_conv2d.run(act_shape, weight_shape,
                                  act_dtype, weight_dtype,
                                  stride, padding,
                                  with_batchnorm, act_func, disable_fusion,
                                  par_ich, par_och, par_col, par_row,
                                  concur_och, stationary,
                                  chunk_size,
                                  axi_datawidth, silent=False,
                                  filename='tmp.v',
                                  outputfile=os.path.splitext(os.path.basename(__file__))[0] + '.out')
    print(rslt)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import math
from collections import OrderedDict

import veriloggen as vg

import nngen.basic_types as bt
import nngen.util as util


class upsampling2d(bt._ElementwiseOperator):
    input_chainable = True
    output_chainable = False

    @staticmethod
    def op(strm, *args, **kwargs):
        return args[0]

    def __init__(self, value, factors, dtype=None, name=None, par=1):

        if len(value.shape) != 4:
            raise ValueError('rank of value.shape must be 4.')

        if len(factors) != 4:
            raise ValueError('rank of factors must be 4.')

        if factors[0] != 1 or factors[3] != 1:
            raise ValueError('factors[0] and [3] must be 1')

        shape = [s * f for s, f in zip(value.shape, factors)]

        bt._ElementwiseOperator.__init__(self, value,
                                         dtype=dtype, shape=shape, name=name, par=par)
        self.factors = factors

    def get_control_param_values(self):
        orig_shape = self.args[0].shape

        num_words = self.get_word_alignment()

        aligned_shape = []
        for s in orig_shape[:-1]:
            aligned_shape.append(s)

        res = num_words - orig_shape[-1] % num_words

        if res == num_words:
            res = 0

        aligned_shape.append(orig_shape[-1] + res)

        aligned_length = bt.shape_to_length(aligned_shape)

        total_size = int(math.ceil(aligned_length / self.par))
        dma_size = int(math.ceil(aligned_shape[-1] / self.par))
        num_comp = int(math.ceil(total_size / dma_size))

        base = bt.to_byte(bt.align_word(orig_shape[-1], self.get_word_alignment()) *
                          self.get_ram_width())

        factor_col = self.factors[2]
        factor_row = self.factors[1]

        out_col_step = base
        out_row_step = base * (self.shape[-2] - (factor_col - 1))
        max_out_pos_col = factor_col - 1
        max_out_pos_row = factor_row - 1

        out_col_inc = base * factor_col
        out_row_inc = out_col_inc + base * self.shape[-2] * (factor_row - 1)
        max_out_col_count = orig_shape[-2] - 1

        sources = self.collect_sources()

        arg_addr_incs = []
        wrap_modes = []
        wrap_sizes = []
        for arg in sources:
            arg_addr_inc = bt.to_byte(bt.align_word(arg.shape[-1], arg.get_word_alignment()) *
                                      arg.get_ram_width())
            if tuple(arg.shape) == tuple(orig_shape):
                wrap_mode = 0
                wrap_size = 0
            elif len(arg.shape) == 1 and arg.shape[-1] == 1:
                # stride-0
                wrap_mode = 2
                wrap_size = bt.get_wrap_size(orig_shape, arg.shape)
            else:
                # repeat
                wrap_mode = 1
                wrap_size = bt.get_wrap_size(orig_shape, arg.shape)
            arg_addr_incs.append(arg_addr_inc)
            wrap_modes.append(wrap_mode)
            wrap_sizes.append(wrap_size)

        return OrderedDict([('dma_size', dma_size),
                            ('num_comp', num_comp),
                            ('out_col_step', out_col_step),
                            ('out_row_step', out_row_step),
                            ('max_out_pos_col', max_out_pos_col),
                            ('max_out_pos_row', max_out_pos_row),
                            ('out_col_inc', out_col_inc),
                            ('out_row_inc', out_row_inc),
                            ('max_out_col_count', max_out_col_count),
                            ('arg_addr_incs', arg_addr_incs),
                            ('wrap_modes', wrap_modes),
                            ('wrap_sizes', wrap_sizes)])

    def control_sequence(self, fsm):
        sources = self.collect_sources()

        arg_gaddrs = [self.m.Reg(self._name('arg_gaddr_%d' % i),
                                 self.maxi.addrwidth, initval=0)
                      for i, _ in enumerate(self.arg_objaddrs)]
        out_gaddr = self.m.Reg(self._name('out_gaddr'),
                               self.maxi.addrwidth, initval=0)
        out_gaddr_offset = self.m.Reg(self._name('out_gaddr_offset'),
                                      self.maxi.addrwidth, initval=0)
        out_pos_col = self.m.Reg(self._name('out_pos_col'),
                                 self.maxi.addrwidth + 1, initval=0)
        out_pos_row = self.m.Reg(self._name('out_pos_row'),
                                 self.maxi.addrwidth + 1, initval=0)
        out_col_count = self.m.Reg(self._name('out_col_count'),
                                   self.maxi.addrwidth + 1, initval=0)
        comp_count = self.m.Reg(self._name('comp_count'),
                                self.maxi.addrwidth + 1, initval=0)
        wrap_counts = [self.m.Reg(self._name('wrap_count_%d' % i),
                                  self.maxi.addrwidth + 1, initval=0)
                       for i, arg in enumerate(sources)]

        arg_pages = [self.m.Reg(self._name('arg_page_%d' % i), initval=0)
                     for i, _ in enumerate(self.arg_objaddrs)]
        arg_page_comp_offsets = [self.m.Reg(self._name('arg_page_comp_offset_%d' % i),
                                            self.maxi.addrwidth, initval=0)
                                 for i, _ in enumerate(self.arg_objaddrs)]
        arg_page_dma_offsets = [self.m.Reg(self._name('arg_page_dma_offset_%d' % i),
                                           self.maxi.addrwidth, initval=0)
                                for i, _ in enumerate(self.arg_objaddrs)]

        out_page = self.m.Reg(self._name('out_page'), initval=0)
        out_page_comp_offset = self.m.Reg(self._name('out_page_comp_offset'),
                                          self.maxi.addrwidth, initval=0)
        out_page_dma_offset = self.m.Reg(self._name('out_page_dma_offset'),
                                         self.maxi.addrwidth, initval=0)

        arg_page_size = self.output_rams[0].length // 2
        out_page_size = self.output_rams[0].length // 2

        skip_read = self.m.Reg(self._name('skip_read'), initval=0)
        skip_comp = self.m.Reg(self._name('skip_comp'), initval=0)
        skip_write = self.m.Reg(self._name('skip_write'), initval=0)

        # --------------------
        # initialization phase
        # --------------------
        fsm(
            [arg_gaddr(0) for arg_gaddr in arg_gaddrs]
        )

        fsm(
            comp_count(0),
            out_gaddr(0),
            out_gaddr_offset(0),
            out_pos_col(0),
            out_pos_row(0),
            out_col_count(0),
            [wrap_count(0) for wrap_count in wrap_counts]
        )

        fsm(
            [arg_page(0) for arg_page in arg_pages],
            [arg_page_comp_offset(0)
             for arg_page_comp_offset in arg_page_comp_offsets],
            [arg_page_dma_offset(0)
             for arg_page_dma_offset in arg_page_dma_offsets]
        )

        fsm(
            out_page(0),
            out_page_comp_offset(0),
            out_page_dma_offset(out_page_size)
        )

        fsm(
            skip_read(0),
            skip_comp(0),
            skip_write(1)
        )

        fsm.goto_next()

        # --------------------
        # Read phase
        # --------------------
        state_read = fsm.current
        fsm.goto_next()

        # DMA read -> Stream run -> Stream wait -> DMA write
        for (ram, arg_objaddr,
             arg_gaddr, arg_page_dma_offset,
             wrap_mode, wrap_count, arg) in zip(self.input_rams, self.arg_objaddrs,
                                                arg_gaddrs, arg_page_dma_offsets,
                                                self.wrap_modes, wrap_counts, sources):

            b = fsm.current
            fsm.goto_next()

            # normal
            laddr = arg_page_dma_offset
            gaddr = arg_objaddr + arg_gaddr
            bt.bus_lock(self.maxi, fsm)
            bt.dma_read(self.maxi, fsm, ram, laddr, gaddr, self.dma_size)
            bt.bus_unlock(self.maxi, fsm)
            fsm.goto_next()

            b_stride0 = fsm.current
            fsm.goto_next()

            # stride-0
            lsize = vg.Int(1)
            bt.bus_lock(self.maxi, fsm)
            bt.dma_read(self.maxi, fsm, ram, laddr, gaddr, lsize)
            bt.bus_unlock(self.maxi, fsm)
            fsm.goto_next()

            # for reuse
            e = fsm.current
            fsm.If(wrap_mode == 2, wrap_count > 0).goto_from(b, e)
            fsm.If(wrap_mode == 2, wrap_count == 0).goto_from(b, b_stride0)
            fsm.If(wrap_mode != 2).goto_from(b_stride0, e)

        state_read_end = fsm.current
        fsm.If(skip_read).goto_from(state_read, state_read_end)

        # --------------------
        # Comp phase
        # --------------------
        state_comp = fsm.current

        self.stream.source_join(fsm)

        # set_source, set_parameter (dup)
        for (source_name, dup_name,
             arg_page_comp_offset,
             ram, wrap_mode) in zip(self.stream.sources.keys(),
                                    self.stream.parameters.keys(),
                                    arg_page_comp_offsets,
                                    self.input_rams, self.wrap_modes):
            read_laddr = arg_page_comp_offset
            read_size = self.dma_size
            stride = vg.Mux(wrap_mode == 2, 0, 1)
            dup = vg.Mux(wrap_mode == 2, 1, 0)
            self.stream.set_parameter(fsm, dup_name, dup)
            fsm.set_index(fsm.current - 1)
            self.stream.set_source(fsm, source_name, ram,
                                   read_laddr, read_size, stride)
            fsm.set_index(fsm.current - 1)

        # set_sink
        write_laddr = out_page_comp_offset
        write_size = self.dma_size

        for name, ram in zip(self.stream.sinks.keys(), self.output_rams):
            self.stream.set_sink(fsm, name, ram, write_laddr, write_size)
            fsm.set_index(fsm.current - 1)

        fsm.goto_next()

        # waiting for previous DMA write completion
        bt.dma_wait_write_idle(self.maxi, fsm)

        self.stream.run(fsm)

        state_comp_end = fsm.current

        self.stream.join(fsm)

        state_comp_end_join = fsm.current

        fsm.If(skip_comp).goto_from(state_comp, state_comp_end)
        fsm.If(vg.Not(skip_comp)).goto_from(
            state_comp_end, state_comp_end_join)

        # --------------------
        # Write phase
        # --------------------
        state_write = fsm.current
        fsm.goto_next()

        laddr = out_page_dma_offset
        gaddr_base = self.objaddr + out_gaddr

        bt.bus_lock(self.maxi, fsm)

        b = fsm.current

        gaddr = gaddr_base + out_gaddr_offset
        bt.dma_write(self.maxi, fsm,
                     self.output_rams[0], laddr, gaddr, self.dma_size, use_async=True)

        fsm(
            out_pos_col.inc(),
            out_gaddr_offset.add(self.out_col_step),
        )
        fsm.If(out_pos_col == self.max_out_pos_col)(
            out_pos_col(0),
            out_pos_row.inc(),
            out_gaddr_offset.add(self.out_row_step)
        )

        fsm.goto(b)
        fsm.If(out_pos_col == self.max_out_pos_col,
               out_pos_row == self.max_out_pos_row).goto_next()

        bt.bus_unlock(self.maxi, fsm)

        fsm.goto_next()

        state_write_end = fsm.current
        fsm.If(skip_write).goto_from(state_write, state_write_end)

        # --------------------
        # update for next iteration
        # --------------------
        fsm(
            comp_count.inc()
        )

        fsm(
            out_gaddr_offset(0),
            out_pos_col(0),
            out_pos_row(0)
        )

        fsm.If(vg.Not(skip_write))(
            out_gaddr.add(self.out_col_inc),
            out_col_count.inc(),
        )
        fsm.If(vg.Not(skip_write), out_col_count == self.max_out_col_count)(
            out_gaddr.add(self.out_row_inc),
            out_col_count(0),
        )

        for (arg_gaddr, arg_addr_inc,
             arg_page, arg_page_comp_offset, arg_page_dma_offset,
             wrap_mode, wrap_size,
             wrap_count, arg) in zip(arg_gaddrs, self.arg_addr_incs,
                                     arg_pages, arg_page_comp_offsets,
                                     arg_page_dma_offsets,
                                     self.wrap_modes, self.wrap_sizes,
                                     wrap_counts, sources):

            fsm.If(wrap_mode == 2)(
                wrap_count(1)
            )

            fsm.If(wrap_mode == 1)(
                arg_gaddr.add(arg_addr_inc),
                wrap_count.inc()
            )
            fsm.If(wrap_mode == 1, wrap_count == wrap_size - 1)(
                arg_gaddr(0),
                wrap_count(0)
            )

            fsm.If(wrap_mode == 0)(
                arg_gaddr.add(arg_addr_inc)
            )

            fsm.If(vg.Not(arg_page), wrap_mode != 2)(
                arg_page_comp_offset(arg_page_size),
                arg_page_dma_offset(out_page_size),
                arg_page(1)
            )
            fsm.If(arg_page, wrap_mode != 2)(
                arg_page_comp_offset(0),
                arg_page_dma_offset(0),
                arg_page(0)
            )

        fsm.If(vg.Not(out_page))(
            out_page_comp_offset(out_page_size),
            out_page_dma_offset(0),
            out_page(1)
        )
        fsm.If(out_page)(
            out_page_comp_offset(0),
            out_page_dma_offset(out_page_size),
            out_page(0)
        )

        fsm(
            skip_write(0)
        )
        fsm.If(comp_count == self.num_comp - 1)(
            skip_read(1),
            skip_comp(1)
        )

        fsm.If(comp_count < self.num_comp).goto(state_read)
        fsm.If(comp_count == self.num_comp).goto_next()

        # wait for last DMA write
        bt.dma_wait_write(self.maxi, fsm)

    def eval(self, memo, input_dict, **kwargs):
        if id(self) in memo:
            return memo[id(self)]

        import nngen.verify as verify

        name = self.__class__.__name__

        args = [arg.eval(memo, input_dict)
                for arg in self.args]

        value = args[0]
        factors = self.factors

        kwargs['factors'] = self.factors
        kwargs['dtype'] = self.dtype
        kwargs['name'] = self.name
        kwargs['par'] = self.par

        method = self.get_eval_method()
        ret = method(value, **kwargs)
        memo[id(self)] = ret

        return ret



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np

import nngen.basic_types as bt
import nngen.operator as operator
import nngen.storage as storage


def get_name(node):
    name = node.name
    if name == '':  # if isinstance(node, onnx.FOO.BAR):
        name = '_'.join([output for output in node.output])
    return name


def get_output_names(node):
    return [name for name in node.output]


def search_node_from_model(model, name):
    for node in model.graph.node:
        output_names = get_output_names(node)

        for output_name in output_names:
            if name == output_name:
                return node

    return None


def to_shape(node, value_shapes):
    shape = tuple([d.dim_value for d in node.type.tensor_type.shape.dim])
    if node.name in value_shapes:
        shape = value_shapes[node.name]
    return shape


def transpose_layout(value, expected_layout, onnx_layout):

    if value.get_layout() == expected_layout:
        return value

    if isinstance(value, bt._Storage) and value.get_layout() is None and not value.consumers:
        if len(expected_layout) != len(onnx_layout):
            raise ValueError('layout format size mismatch: %d != %d' %
                             (len(expected_layout), len(onnx_layout)))

        perm = tuple([onnx_layout.index(e) for e in expected_layout])
        new_shape = tuple([value.shape[p] for p in perm])

        value.shape = new_shape
        value.layout = expected_layout
        value.onnx_layout = onnx_layout
        value.perm = perm

        if value.value is not None:
            value.value = np.transpose(value.value, perm)

        return value

    current_layout = value.get_layout()

    if current_layout == expected_layout:
        return value

    if current_layout is None:
        current_layout = onnx_layout

    perm = tuple([current_layout.index(e) for e in expected_layout])

    value = operator.transpose(value, perm)
    value.implicit = True
    value.layout = expected_layout
    value.onnx_layout = onnx_layout

    return value


def convert_transpose_perm(perm, src_layout, dst_layout):
    return tuple([dst_layout.index(src_layout[p]) for p in perm])


def optimize_to_raw_value(value):

    while True:
        if isinstance(value, bt._View) and value.dtype == value.args[0].dtype:
            value = value.args[0]
            continue

        if isinstance(value, bt._LazyReshape) and not value.args[0].shape:
            value = value.args[0]
            continue

        break

    if isinstance(value, storage.variable):
        value = value.value
        return value

    if isinstance(value, storage.constant):
        value = value.value
        return value

    return value


def infer_shape(shape, length):

    if not isinstance(shape, list):
        shape = list(shape)

    use_minus_one = False
    minus_one_index = 0
    all_mul = 1

    for i, s in enumerate(shape):
        if s is None or s == -1:
            use_minus_one = True
            minus_one_index = i
        elif s < 0:
            raise ValueError('not supported value for shape: %d' % s)
        else:
            all_mul *= s

    if use_minus_one:
        shape[minus_one_index] = length // all_mul

    shape = tuple(shape)
    return shape



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import collections

import nngen.operator as operator
import nngen.dtype_list as dtype_list

from . import util


def _pool(pool_op, visitor, node, no_sum_dtype=False):

    srcs = []

    for src in node.input:
        src_obj = visitor.visit(src)
        srcs.append(src_obj)

    input = srcs[0]

    # transpose data layout to nngen-compatible format
    input = util.transpose_layout(input, visitor.nngen_input_layout, visitor.onnx_input_layout)

    name = util.get_name(node)

    if name in visitor.value_dtypes:
        dtype = visitor.value_dtypes[name]
    else:
        dtype = visitor.default_operator_dtype

    if dtype.width >= 16:
        sum_dtype = dtype_list.dtype_int(dtype.width * 4)
    else:
        sum_dtype = dtype_list.int32

    ksize, strides, padding = _get_ksize_strides_padding(node)

    kwargs = collections.OrderedDict()
    kwargs['ksize'] = ksize
    kwargs['strides'] = strides
    kwargs['padding'] = padding
    kwargs['dtype'] = dtype
    if not no_sum_dtype:
        kwargs['sum_dtype'] = sum_dtype
    kwargs['name'] = name

    c = pool_op(input, **kwargs)
    c.layout = visitor.nngen_input_layout
    c.onnx_layout = visitor.onnx_input_layout

    return c


def AveragePool(visitor, node):

    ksize, strides, padding = _get_ksize_strides_padding(node)

    if ksize == strides:
        op = operator.avg_pool_serial
    else:
        op = operator.avg_pool

    return _pool(op, visitor, node)


def GlobalAveragePool(visitor, node):

    srcs = []

    for src in node.input:
        src_obj = visitor.visit(src)
        srcs.append(src_obj)

    input = srcs[0]

    # transpose data layout to nngen-compatible format
    input = util.transpose_layout(input, visitor.nngen_input_layout, visitor.onnx_input_layout)

    name = util.get_name(node)

    if name in visitor.value_dtypes:
        dtype = visitor.value_dtypes[name]
    else:
        dtype = visitor.default_operator_dtype

    if dtype.width >= 16:
        sum_dtype = dtype_list.dtype_int(dtype.width * 4)
    else:
        sum_dtype = dtype_list.int32

    ksize = tuple([1, input.shape[1], input.shape[2], 1])
    strides = tuple([1, input.shape[1], input.shape[2], 1])
    padding = tuple([0, 0, 0, 0])

    kwargs = collections.OrderedDict()
    kwargs['ksize'] = ksize
    kwargs['strides'] = strides
    kwargs['padding'] = padding
    kwargs['dtype'] = dtype
    kwargs['name'] = name

    c = operator.avg_pool_serial(input, **kwargs)
    c.layout = visitor.nngen_input_layout

    return c


def MaxPool(visitor, node):

    ksize, strides, padding = _get_ksize_strides_padding(node)

    if ksize == strides:
        op = operator.max_pool_serial
    else:
        op = operator.max_pool

    return _pool(op, visitor, node, no_sum_dtype=True)


def _get_ksize_strides_padding(node):

    ksize = [1, 1, 1, 1]  # B, H, W, C
    strides = [1, 1, 1, 1]  # B, H, W, C
    padding = [0, 0, 0, 0]  # Top, Bottom, Left, Right

    for attribute in node.attribute:
        if attribute.name == 'auto_pad':
            padding = 'SAME'

        elif attribute.name == 'pads':
            padding[0] = attribute.ints[0]
            padding[1] = attribute.ints[1]
            padding[2] = attribute.ints[2]
            padding[3] = attribute.ints[3]
            padding = tuple(padding)

        elif attribute.name == 'strides':
            strides[1] = attribute.ints[0]
            strides[2] = attribute.ints[1]
            strides = tuple(strides)

        elif attribute.name == 'kernel_shape':
            ksize[1] = attribute.ints[0]
            ksize[2] = attribute.ints[1]
            ksize = tuple(ksize)

    return ksize, strides, padding



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division


from . import conv2d


def matmul(visitor, node):

    return conv2d.conv2d(visitor, node)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np

import nngen.util as util


def add(x, y, dtype=None, name=None, par=1,
        x_dtype=None, y_dtype=None):

    x_point = 0 if x_dtype is None else x_dtype.point
    y_point = 0 if y_dtype is None else y_dtype.point
    xy_point = max(x_point, y_point)
    out_point = xy_point if dtype is None else dtype.point
    out_shift = out_point - xy_point

    x = x << (out_point - x_point)
    y = y << (out_point - y_point)

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    ret = out_op(x + y)

    return ret


def sub(x, y, dtype=None, name=None, par=1,
        x_dtype=None, y_dtype=None):

    x_point = 0 if x_dtype is None else x_dtype.point
    y_point = 0 if y_dtype is None else y_dtype.point
    xy_point = max(x_point, y_point)
    out_point = xy_point if dtype is None else dtype.point
    out_shift = out_point - xy_point

    x = x << (out_point - x_point)
    y = y << (out_point - y_point)

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    ret = out_op(x - y)

    return ret


def neg(x, dtype=None, name=None, par=1,
        x_dtype=None):

    x_point = 0 if x_dtype is None else x_dtype.point
    out_point = x_point if dtype is None else dtype.point
    out_shift = out_point - x_point

    x = x << (out_point - x_point)

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    ret = out_op(np.negative(x))

    return ret


def abs(x, dtype=None, name=None, par=1,
        x_dtype=None):

    x_point = 0 if x_dtype is None else x_dtype.point
    out_point = x_point if dtype is None else dtype.point
    out_shift = out_point - x_point

    x = x << (out_point - x_point)

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    ret = out_op(np.abs(x))

    return ret


def zeros_imm(shape, dtype=None, name=None, par=1):

    out_point = 0 if dtype is None else dtype.point
    out_shift = out_point

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    ret = out_op(np.zeros(shape, dtype=np.int64))

    return ret


def zeros_imm_like(x, dtype=None, name=None, par=1):
    shape = x.shape
    return zeros_imm(shape, dtype, name, par)


def ones_imm(shape, dtype=None, name=None, par=1):

    out_point = 0 if dtype is None else dtype.point
    out_shift = out_point

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    ret = out_op(np.ones(shape, dtype=np.int64))

    return ret


def ones_imm_like(x, dtype=None, name=None, par=1):
    shape = x.shape
    return ones_imm(shape, dtype, name, par)


def full_imm(shape, fill_value, dtype=None, name=None, par=1):

    out_point = 0 if dtype is None else dtype.point
    out_shift = out_point

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    ret = out_op(np.full(shape, fill_value, dtype=np.int64))

    return ret


def full_imm_like(x, fill_value, dtype=None, name=None, par=1):
    shape = x.shape
    return full_imm(shape, fill_value, dtype, name, par)


def equal(x, y, dtype=None, name=None, par=1,
          x_dtype=None, y_dtype=None):

    x_point = 0 if x_dtype is None else x_dtype.point
    y_point = 0 if y_dtype is None else y_dtype.point
    xy_point = max(x_point, y_point)
    out_point = 0 if dtype is None else dtype.point
    out_shift = out_point - xy_point

    x = x << (out_point - x_point)
    y = y << (out_point - y_point)

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    ret = out_op(x == y)

    return ret


def not_equal(x, y, dtype=None, name=None, par=1,
              x_dtype=None, y_dtype=None):

    x_point = 0 if x_dtype is None else x_dtype.point
    y_point = 0 if y_dtype is None else y_dtype.point
    xy_point = max(x_point, y_point)
    out_point = 0 if dtype is None else dtype.point
    out_shift = out_point - xy_point

    x = x << (out_point - x_point)
    y = y << (out_point - y_point)

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    ret = out_op(x != y)

    return ret


def less(x, y, dtype=None, name=None, par=1,
         x_dtype=None, y_dtype=None):

    x_point = 0 if x_dtype is None else x_dtype.point
    y_point = 0 if y_dtype is None else y_dtype.point
    xy_point = max(x_point, y_point)
    out_point = 0 if dtype is None else dtype.point
    out_shift = out_point - xy_point

    x = x << (out_point - x_point)
    y = y << (out_point - y_point)

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    ret = out_op(x < y)

    return ret


def less_equal(x, y, dtype=None, name=None, par=1,
               x_dtype=None, y_dtype=None):

    x_point = 0 if x_dtype is None else x_dtype.point
    y_point = 0 if y_dtype is None else y_dtype.point
    xy_point = max(x_point, y_point)
    out_point = 0 if dtype is None else dtype.point
    out_shift = out_point - xy_point

    x = x << (out_point - x_point)
    y = y << (out_point - y_point)

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    ret = out_op(x <= y)

    return ret


def greater(x, y, dtype=None, name=None, par=1,
            x_dtype=None, y_dtype=None):

    x_point = 0 if x_dtype is None else x_dtype.point
    y_point = 0 if y_dtype is None else y_dtype.point
    xy_point = max(x_point, y_point)
    out_point = 0 if dtype is None else dtype.point
    out_shift = out_point - xy_point

    x = x << (out_point - x_point)
    y = y << (out_point - y_point)

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    ret = out_op(x > y)

    return ret


def greater_equal(x, y, dtype=None, name=None, par=1,
                  x_dtype=None, y_dtype=None):

    x_point = 0 if x_dtype is None else x_dtype.point
    y_point = 0 if y_dtype is None else y_dtype.point
    xy_point = max(x_point, y_point)
    out_point = 0 if dtype is None else dtype.point
    out_shift = out_point - xy_point

    x = x << (out_point - x_point)
    y = y << (out_point - y_point)

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    ret = out_op(x >= y)

    return ret


def sign_binary(x, dtype=None, name=None, par=1,
                x_dtype=None):

    x_point = 0 if x_dtype is None else x_dtype.point
    out_point = x_point if dtype is None else dtype.point
    out_shift = out_point - x_point

    x = x << (out_point - x_point)

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    ret = out_op(np.where(x > np.zeros_like(x, dtype=np.int64),
                          np.ones_like(x, dtype=np.int64),
                          np.negative(np.ones_like(x, dtype=np.int64))))

    return ret


def sign_ternary(x, dtype=None, name=None, par=1,
                 x_dtype=None):

    x_point = 0 if x_dtype is None else x_dtype.point
    out_point = x_point if dtype is None else dtype.point
    out_shift = out_point - x_point

    x = x << (out_point - x_point)

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    ret = out_op(np.where(x > np.zeros_like(x, dtype=np.int64),
                          np.ones_like(x, dtype=np.int64),
                          np.where(x == np.zeros_like(x, dtype=np.int64),
                                   np.zeros_like(x, dtype=np.int64),
                                   np.negative(np.ones_like(x, dtype=np.int64)))))

    return ret


def where(condition, x, y, dtype=None, name=None, par=1,
          condition_dtype=None, x_dtype=None, y_dtype=None):

    x_point = 0 if x_dtype is None else x_dtype.point
    y_point = 0 if y_dtype is None else y_dtype.point
    xy_point = max(x_point, y_point)
    out_point = xy_point if dtype is None else dtype.point
    out_shift = out_point - xy_point

    x = x << (out_point - x_point)
    y = y << (out_point - y_point)

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    ret = out_op(np.where(condition, x, y))

    return ret


def add_n(arg, dtype=None, name=None, par=1,
          arg_dtypes=None):

    if not isinstance(arg, (tuple, list)):
        raise TypeError('expected tuple or list')

    if arg_dtypes is None:
        arg_dtypes = [None for _ in arg]

    if not isinstance(arg_dtypes, (tuple, list)):
        raise TypeError('expected tuple or list')

    if len(arg) != len(arg_dtypes):
        raise ValueError('length mismatch: %d != %d' % (len(arg), len(arg_dtypes)))

    arg_points = [0 if a_dtype is None else a_dtype.point
                  for a_dtype in arg_dtypes]
    max_arg_point = max(*arg_points)

    out_point = max_arg_point if dtype is None else dtype.point
    out_shift = out_point - max_arg_point

    values = [a << (out_point - a_point)
              for a, a_point in zip(arg, arg_points)]

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    ret = values[0]
    for value in values[1:]:
        ret += value

    ret = out_op(ret)

    return ret


def lshift(x, y, dtype=None, name=None, par=1,
           x_dtype=None, y_dtype=None):

    x_point = 0 if x_dtype is None else x_dtype.point
    y_point = 0 if y_dtype is None else y_dtype.point
    xy_point = max(x_point, y_point)
    xy_shift = min(x_point, y_point)
    out_point = xy_point if dtype is None else dtype.point
    out_shift = -xy_shift if dtype is None else dtype.point - xy_point - xy_shift

    x = x << (out_point - x_point)
    y = y << (out_point - y_point)

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    ret = out_op(x << y)

    return ret


def rshift(x, y, dtype=None, name=None, par=1,
           x_dtype=None, y_dtype=None):

    x_point = 0 if x_dtype is None else x_dtype.point
    y_point = 0 if y_dtype is None else y_dtype.point
    xy_point = max(x_point, y_point)
    xy_shift = min(x_point, y_point)
    out_point = xy_point if dtype is None else dtype.point
    out_shift = -xy_shift if dtype is None else dtype.point - xy_point - xy_shift

    x = x << (out_point - x_point)
    y = y << (out_point - y_point)

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    ret = out_op(x >> y)

    return ret


def rshift_round(x, y, dtype=None, name=None, par=1,
                 x_dtype=None, y_dtype=None):

    x_point = 0 if x_dtype is None else x_dtype.point
    y_point = 0 if y_dtype is None else y_dtype.point
    xy_point = max(x_point, y_point)
    xy_shift = min(x_point, y_point)
    out_point = xy_point if dtype is None else dtype.point
    out_shift = -xy_shift if dtype is None else dtype.point - xy_point - xy_shift

    x = x << (out_point - x_point)
    y = y << (out_point - y_point)

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    shifted = x >> y
    last_bit = (x >> (y - 1)) & 0x1

    ret = out_op(np.where(y == 0, x, shifted + last_bit))

    return ret


def clip(x, asymmetric_clip=False,
         dtype=None, name=None, par=1,
         x_dtype=None):

    x_point = 0 if x_dtype is None else x_dtype.point
    out_point = 0 if dtype is None else dtype.point
    out_shift = out_point - x_point

    width = 32 if dtype is None else dtype.width
    signed = True if dtype is None else dtype.signed

    p_th, n_th = util.clip_threshold(width, signed, asymmetric_clip)
    p_th = np.ones_like(x, dtype=np.int64) * [p_th]
    n_th = np.ones_like(x, dtype=np.int64) * [n_th]

    p = np.where(x > p_th, p_th, x)
    n = np.where(x < n_th, n_th, x)
    x = np.where(x >= 0, p, n)

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    ret = out_op(x)

    return ret


def multiply(x, y, dtype=None, name=None, par=1,
             x_dtype=None, y_dtype=None):

    x_point = 0 if x_dtype is None else x_dtype.point
    y_point = 0 if y_dtype is None else y_dtype.point
    xy_point = max(x_point, y_point)
    xy_shift = min(x_point, y_point)
    out_point = xy_point if dtype is None else dtype.point
    out_shift = -xy_shift if dtype is None else dtype.point - xy_point - xy_shift

    x = x << (out_point - x_point)
    y = y << (out_point - y_point)

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    ret = out_op(x * y)

    return ret


def multiply_shared(x, y, dtype=None, name=None, par=1,
                    x_dtype=None, y_dtype=None):

    return multiply(x, y, dtype, name, par, x_dtype, y_dtype)


def div(x, y, dtype=None, name=None, par=1,
        x_dtype=None, y_dtype=None):

    x_point = 0 if x_dtype is None else x_dtype.point
    y_point = 0 if y_dtype is None else y_dtype.point
    xy_point = max(x_point, y_point)
    xy_shift = min(x_point, y_point)
    out_point = xy_point if dtype is None else dtype.point
    out_shift = -xy_shift if dtype is None else dtype.point - xy_point - xy_shift

    x = x << (out_point - x_point)
    y = y << (out_point - y_point)

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    ret = out_op(x // y)

    return ret


def multiply_add_rshift_clip(x, y, z, shamt, asymmetric_clip=False,
                             dtype=None, sum_dtype=None, name=None, par=1,
                             x_dtype=None, y_dtype=None, z_dtype=None, shamt_dtype=None):

    v = multiply(x, y, dtype=sum_dtype, par=par,
                 x_dtype=x_dtype, y_dtype=y_dtype)
    v = add(v, z, dtype=sum_dtype, par=par,
            x_dtype=sum_dtype, y_dtype=z_dtype)
    v = rshift(v, shamt, dtype=sum_dtype, par=par,
               x_dtype=sum_dtype, y_dtype=shamt_dtype)
    return clip(v, asymmetric_clip=asymmetric_clip,
                dtype=dtype, par=par,
                x_dtype=sum_dtype)


def reduce_sum(input_tensor,
               axis=None, keep_dims=False, dtype=None, name=None, par=1,
               input_tensor_dtype=None):

    input_tensor_point = 0 if input_tensor_dtype is None else input_tensor_dtype.point
    out_point = input_tensor_point if dtype is None else dtype.point
    out_shift = out_point - input_tensor_point

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    if axis is None:
        input_tensor = input_tensor.reshape([-1])

    ret = np.sum(input_tensor, axis, keepdims=keep_dims)

    if not isinstance(ret, np.ndarray):
        ret = np.array([ret])

    ret = out_op(ret)

    return ret


def reduce_max(input_tensor,
               axis=None, keep_dims=False, dtype=None, name=None, par=1,
               input_tensor_dtype=None):

    input_tensor_point = 0 if input_tensor_dtype is None else input_tensor_dtype.point
    out_point = input_tensor_point if dtype is None else dtype.point
    out_shift = out_point - input_tensor_point

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    if axis is None:
        input_tensor = input_tensor.reshape([-1])

    ret = np.max(input_tensor, axis=axis, keepdims=keep_dims)

    if not isinstance(ret, np.ndarray):
        ret = np.array([ret])

    ret = out_op(ret)

    return ret


def reduce_min(input_tensor,
               axis=None, keep_dims=False, dtype=None, name=None, par=1,
               input_tensor_dtype=None):

    input_tensor_point = 0 if input_tensor_dtype is None else input_tensor_dtype.point
    out_point = input_tensor_point if dtype is None else dtype.point
    out_shift = out_point - input_tensor_point

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    if axis is None:
        input_tensor = input_tensor.reshape([-1])

    ret = np.min(input_tensor, axis=axis, keepdims=keep_dims)

    if not isinstance(ret, np.ndarray):
        ret = np.array([ret])

    ret = out_op(ret)

    return ret


def argmax(input_tensor,
           axis=None, keep_dims=False, dtype=None, name=None, par=1,
           input_tensor_dtype=None):

    input_tensor_point = 0 if input_tensor_dtype is None else input_tensor_dtype.point
    out_point = input_tensor_point if dtype is None else dtype.point
    out_shift = out_point - input_tensor_point

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    if isinstance(axis, (tuple, list, np.ndarray)) and len(axis) > 1:
        raise ValueError('Size of axis must be 1.')

    if isinstance(axis, (tuple, list, np.ndarray)):
        axis = int(axis[0])

    if axis is None:
        input_tensor = input_tensor.reshape([-1])

    ret = np.argmax(input_tensor, axis=axis, keepdims=keep_dims)

    if not isinstance(ret, np.ndarray):
        ret = np.array([ret])

    ret = out_op(ret)

    return ret


def argmin(input_tensor,
           axis=None, keep_dims=False, dtype=None, name=None, par=1,
           input_tensor_dtype=None):

    input_tensor_point = 0 if input_tensor_dtype is None else input_tensor_dtype.point
    out_point = input_tensor_point if dtype is None else dtype.point
    out_shift = out_point - input_tensor_point

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    if isinstance(axis, (tuple, list, np.ndarray)) and len(axis) > 1:
        raise ValueError('Size of axis must be 1.')

    if isinstance(axis, (tuple, list, np.ndarray)):
        axis = int(axis[0])

    if axis is None:
        input_tensor = input_tensor.reshape([-1])

    ret = np.argmin(input_tensor, axis=axis, keepdims=keep_dims)

    if not isinstance(ret, np.ndarray):
        ret = np.array([ret])

    ret = out_op(ret)

    return ret


def _reshape(tensor, shape, dtype=None, name=None, tensor_dtype=None):
    return np.reshape(tensor, shape)


def _lazy_reshape(tensor, shape, dtype=None, name=None, tensor_dtype=None):
    return np.reshape(tensor, shape)


def _View(tensor, shape, dtype=None, name=None):
    return np.reshape(tensor, shape)


def reshape(tensor, shape, dtype=None, name=None, tensor_dtype=None):
    return np.reshape(tensor, shape)


def cast(x, dtype=None, name=None):
    return x


def expand_dims(input, axis, name=None):
    shape = input.shape
    rank = len(shape)
    axis = util.to_axis(axis, rank)[0]
    new_shape = list(shape[:axis]) + [1] + list(shape[axis:])

    return input.reshape(new_shape)


def transpose(a, perm=None, dtype=None, name=None, a_dtype=None):
    return np.transpose(a, perm)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from . import verilog


__intrinsics__ = ('set_header', 'get_header',
                  'set_global_offset', 'set_global_addrs',
                  'set_global_addr_map', 'write_global_addr_map', 'load_global_addr_map',
                  'start', 'wait', 'sw_rst')


def set_header(fsm, saxi, index, header, wordsize=4):
    awaddr = (verilog.header_reg + index) * wordsize
    saxi.write(fsm, awaddr, header)


def get_header(fsm, saxi, index, wordsize=4):
    araddr = (verilog.header_reg + index) * wordsize
    h = saxi.read(fsm, araddr)
    return h


def set_global_offset(fsm, saxi, addr, **opt):
    wordsize = opt['wordsize'] if 'wordsize' in opt else 4

    awaddr = verilog.control_reg_global_offset * wordsize
    saxi.write(fsm, awaddr, addr)


def set_global_addrs(fsm, saxi, *addrs, **opt):
    wordsize = opt['wordsize'] if 'wordsize' in opt else 4
    offset = opt['offset'] if 'offset' in opt else 0

    awaddr = (offset + verilog.control_reg_global_addr) * wordsize
    for addr in addrs:
        saxi.write(fsm, awaddr, addr)
        awaddr += wordsize


def set_global_addr_map(fsm, saxi, memory, map_addr, *addrs, **opt):
    write_global_addr_map(fsm, memory, map_addr, *addrs, **opt)
    load_global_addr_map(fsm, saxi, map_addr, **opt)


def write_global_addr_map(fsm, memory, map_addr, *addrs, **opt):
    wordsize = opt['wordsize'] if 'wordsize' in opt else 4
    offset = opt['offset'] if 'offset' in opt else 0

    for i, addr in enumerate(addrs):
        memory.write_word(fsm, i + offset, map_addr, addr, wordsize * 8)


def load_global_addr_map(fsm, saxi, map_addr, **opt):
    wordsize = opt['wordsize'] if 'wordsize' in opt else 4

    awaddr = verilog.control_reg_addr_global_addr_map * wordsize
    saxi.write(fsm, awaddr, map_addr)

    awaddr = verilog.control_reg_load_global_addr_map * wordsize
    saxi.write(fsm, awaddr, 1)

    araddr = verilog.control_reg_load_global_addr_map * wordsize
    b = fsm.current
    v = saxi.read(fsm, araddr)
    fsm.If(v != 0).goto(b)
    fsm.If(v == 0).goto_next()

    araddr = verilog.control_reg_busy_global_addr_map * wordsize
    b = fsm.current
    v = saxi.read(fsm, araddr)
    fsm.If(v != 0).goto(b)
    fsm.If(v == 0).goto_next()


def start(fsm, saxi, wordsize=4):
    awaddr = verilog.control_reg_start * wordsize
    saxi.write(fsm, awaddr, 1)

    araddr = verilog.control_reg_start * wordsize
    b = fsm.current
    v = saxi.read(fsm, araddr)
    fsm.If(v != 0).goto(b)
    fsm.If(v == 0).goto_next()


def wait(fsm, saxi, wordsize=4):
    araddr = verilog.control_reg_busy * wordsize
    b = fsm.current
    v = saxi.read(fsm, araddr)
    fsm.If(v != 0).goto(b)
    fsm.If(v == 0).goto_next()


def sw_rst(fsm, saxi, wordsize=4):
    awaddr = verilog.control_reg_reset * wordsize
    saxi.write(fsm, awaddr, 1)

    araddr = verilog.control_reg_busy * wordsize
    b = fsm.current
    v = saxi.read(fsm, araddr)
    fsm.If(v != 0).goto(b)
    fsm.If(v == 0).goto_next()



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np


def sigmoid(features,
            lut_addrwidth=8, lut_clip=6.0, range_rate=0.95,
            dtype=None, name=None, par=1,
            features_dtype=None, features_scale=1, features_shamt=0):

    features_point = 0 if features_dtype is None else features_dtype.point
    out_point = 0 if dtype is None else dtype.point
    out_shift = out_point - features_point

    mul = features * features_scale
    sra = mul >> features_shamt

    if dtype is None:
        raise ValueError('sigmoid requires dtype to determine the value range.')

    out_width = dtype.width
    out_point = dtype.point
    out_signed = dtype.signed
    if out_signed:
        out_scale = round((2 ** (out_width - 1)) * range_rate)
    else:
        out_scale = round((2 ** out_width) * range_rate)

    def _sigmoid(x):
        return np.around((1 / (1 + np.exp(-x))) * out_scale).astype(np.int64)

    addr_scale = lut_clip / (2 ** (lut_addrwidth - 1))
    lut = _sigmoid(sra * addr_scale)

    p_th = 2 ** (lut_addrwidth - 1) - 1
    n_th = -1 * p_th

    if out_point == 0:
        th_scale = out_scale
    elif out_point > 0:
        th_scale = out_scale >> out_point
    else:
        th_scale = out_scale << (-1 * out_point)

    p = np.where(sra > p_th, th_scale, lut)
    n = np.where(sra < n_th, 0, lut)
    out = np.where(sra >= 0, p, n)

    return out



from __future__ import absolute_import
from __future__ import print_function

import functools

import veriloggen as vg
from veriloggen.optimizer import try_optimize as optimize

import nngen.basic_types as bt
from . import basic
from . import conv2d


def to_shape_2d(shape):
    if bt.get_rank(shape) == 1:
        return tuple([1, shape[0]])

    return shape


def to_conv2d_act_shape(shape):
    bat = 1
    if bt.get_rank(shape) == 1:
        row = 1
    else:
        row = functools.reduce(lambda x, y: x * y, shape[:-1], 1)
    col = 1
    ch = shape[-1]

    return (bat, row, col, ch)


def to_conv2d_weight_shape(shape):
    # expect transposed data
    if bt.get_rank(shape) == 1:
        och = 1
    else:
        och = functools.reduce(lambda x, y: x * y, shape[:-1], 1)
    row = 1
    col = 1
    ich = shape[-1]

    return (och, row, col, ich)


class matmul(conv2d.conv2d):
    input_chainable = False
    output_chainable = False

    @property
    def par_left_col(self):
        return self.par_ich

    @par_left_col.setter
    def par_left_col(self, par_left_col):
        self.par_ich = par_left_col

    @property
    def par_left_row(self):
        return self.par_col

    @par_left_row.setter
    def par_left_row(self, par_left_row):
        self.par_col = par_left_row

    @property
    def par_out_col(self):
        return self.par_och

    @par_out_col.setter
    def par_out_col(self, par_out_col):
        self.par_och = par_out_col

    @property
    def concur_out_col(self):
        return self.concur_och

    @concur_out_col.setter
    def concur_out_col(self, concur_out_col):
        self.concur_och = concur_out_col

    @property
    def left_ram_size(self):
        return self.input_ram_size

    @left_ram_size.setter
    def left_ram_size(self, left_ram_size):
        self.input_ram_size = left_ram_size

    @property
    def right_ram_size(self):
        return self.filter_ram_size

    @right_ram_size.setter
    def right_ram_size(self, right_ram_size):
        self.filter_ram_size = right_ram_size

    def __sub_str__(self):
        bias = (' bias:%s' % str(self.args[self.args_dict['bias']].shape)
                if 'bias' in self.args_dict else '')
        scale = (' scale:%s' % str(self.args[self.args_dict['scale']].shape)
                 if 'scale' in self.args_dict else '')

        vshamt_mul = (' vshamt_mul:%s' % str(self.args[self.args_dict['vshamt_mul']].shape)
                      if 'vshamt_mul' in self.args_dict else '')
        vshamt_sum = (' vshamt_sum:%s' % str(self.args[self.args_dict['vshamt_sum']].shape)
                      if 'vshamt_sum' in self.args_dict else '')
        vshamt_out = (' vshamt_out:%s' % str(self.args[self.args_dict['vshamt_out']].shape)
                      if 'vshamt_out' in self.args_dict else '')

        cshamt_mul = (' cshamt_mul:%s' % self.cshamt_mul
                      if self.cshamt_mul is not None else '')
        cshamt_sum = (' cshamt_sum:%s' % self.cshamt_sum
                      if self.cshamt_sum is not None else '')
        cshamt_out = (' cshamt_out:%s' % self.cshamt_out
                      if self.cshamt_out is not None else '')

        act_func = (' act_func:%s' % str(self.act_func.__class__.__name__)
                    if self.act_func is not None else '')
        mul_dtype = (' mul_dtype:%s' % self.mul_dtype.to_str()
                     if self.mul_dtype is not None else '')
        sum_dtype = (' sum_dtype:%s' % self.sum_dtype.to_str()
                     if self.sum_dtype is not None else '')

        par_left_col = ' par_left_col:%d' % self.par_left_col if self.par_left_col > 1 else ''
        par_left_row = ' par_left_row:%d' % self.par_left_row if self.par_left_row > 1 else ''
        par_out_col = ' par_out_col:%d' % self.par_out_col if self.par_out_col > 1 else ''
        if hasattr(self, 'concur_och_value'):
            concur_out_col = ' concur_out_col:%d' % self.concur_och_value
        else:
            concur_out_col = ' concur_out_col:%s' % str(self.concur_out_col)
        stationary = ' stationary:%s' % (
            'right' if self.stationary == 'filter' else 'left')

        left_ram_size = (' left_ram_size:%d' % self.input_ram_size
                         if self.left_ram_size is not None else '')
        right_ram_size = (' right_ram_size:%d' % self.filter_ram_size
                          if self.right_ram_size is not None else '')
        bias_ram_size = (' bias_ram_size:%d' % self.bias_ram_size
                         if self.bias_ram_size is not None else '')
        scale_ram_size = (' scale_ram_size:%d' % self.scale_ram_size
                          if self.scale_ram_size is not None else '')
        vshamt_mul_ram_size = (' vshamt_mul_ram_size:%d' % self.vshamt_mul_ram_size
                               if self.vshamt_mul_ram_size is not None else '')
        vshamt_sum_ram_size = (' vshamt_sum_ram_size:%d' % self.vshamt_sum_ram_size
                               if self.vshamt_sum_ram_size is not None else '')
        vshamt_out_ram_size = (' vshamt_out_ram_size:%d' % self.vshamt_out_ram_size
                               if self.vshamt_out_ram_size is not None else '')
        out_ram_size = (' out_ram_size:%d' % self.out_ram_size
                        if self.out_ram_size is not None else '')

        if hasattr(self, 'keep_input_value'):
            keep_left = ' keep_left' if self.keep_input_value else ''
        else:
            keep_left = ''
        if hasattr(self, 'keep_filter_value'):
            keep_right = ' keep_right' if self.keep_filter_value else ''
        else:
            keep_right = ''

        ret = [str(s)
               for s in (bias, scale,
                         vshamt_mul, vshamt_sum, vshamt_out,
                         cshamt_mul, cshamt_sum, cshamt_out,
                         act_func, mul_dtype, sum_dtype,
                         par_left_col, par_left_row, par_out_col,
                         concur_out_col, stationary,
                         left_ram_size, right_ram_size,
                         bias_ram_size, scale_ram_size,
                         vshamt_mul_ram_size, vshamt_sum_ram_size, vshamt_out_ram_size,
                         out_ram_size,
                         keep_left, keep_right)]
        return ''.join(ret)

    def __init__(self, a, b,
                 bias=None, scale=None,
                 transposed_a=False, transposed_b=False,
                 rshift_mul=None, rshift_sum=None, rshift_out=None,
                 act_func=None,
                 asymmetric_clip=False,
                 dtype=None, mul_dtype=None, sum_dtype=None,
                 name=None,
                 par_left_col=1, par_left_row=1, par_out_col=1,
                 concur_out_col=None, stationary='right',
                 left_ram_size=None, right_ram_size=None,
                 bias_ram_size=None, scale_ram_size=None,
                 vshamt_mul_ram_size=None, vshamt_sum_ram_size=None, vshamt_out_ram_size=None,
                 out_ram_size=None,
                 disable_keep_left=False):

        if transposed_a:
            perm = list(range(bt.get_rank(to_shape_2d(a.shape))))
            perm[-2], perm[-1] = perm[-1], perm[-2]
            a = basic.transpose(a, perm=perm)

        if not transposed_b:
            # matrix B must be transposed for fast computation
            perm = list(range(bt.get_rank(to_shape_2d(b.shape))))
            perm[-2], perm[-1] = perm[-1], perm[-2]
            b = basic.transpose(b, perm=perm)

        self.transposed_a = transposed_a
        self.transposed_b = transposed_b

        input_shape = to_conv2d_act_shape(a.shape)
        input = a

        filter_shape = to_conv2d_weight_shape(b.shape)
        filter = b

        strides = (1, 1, 1, 1)
        padding = 'SAME'
        out_shape = tuple(list(to_shape_2d(a.shape)[:-2]) +
                          [to_shape_2d(a.shape)[-2], to_shape_2d(b.shape)[-2]])

        if stationary == 'right':
            stationary = 'filter'
        elif stationary == 'left':
            stationary = 'input'
        else:
            raise ValueError("stationary must be 'left' or 'right'")

        conv2d.conv2d.__init__(self, input, filter, strides,
                               bias, scale,
                               rshift_mul, rshift_sum, rshift_out,
                               act_func, padding,
                               asymmetric_clip,
                               dtype, mul_dtype, sum_dtype,
                               name,
                               par_left_col, par_out_col, par_left_row, 1,
                               concur_out_col, stationary,
                               left_ram_size, right_ram_size,
                               bias_ram_size, scale_ram_size,
                               vshamt_mul_ram_size, vshamt_sum_ram_size, vshamt_out_ram_size,
                               out_ram_size,
                               disable_keep_left,
                               input_shape, filter_shape, out_shape)

    def attribute(self, cshamt_mul=None, cshamt_sum=None, cshamt_out=None,
                  par_left_col=None, par_out_col=None, par_left_row=None,
                  concur_out_col=None,
                  stationary=None,
                  left_ram_size=None, right_ram_size=None,
                  bias_ram_size=None, scale_ram_size=None,
                  vshamt_mul_ram_size=None, vshamt_sum_ram_size=None, vshamt_out_ram_size=None,
                  out_ram_size=None,
                  par_ich=None, par_och=None, par_col=None, par_row=None,
                  concur_och=None,
                  input_ram_size=None, filter_ram_size=None,
                  disable_keep_left=None):

        if stationary is None:
            pass
        elif stationary == 'filter' or stationary == 'input':
            pass
        elif stationary == 'right':
            stationary = 'filter'
        elif stationary == 'left':
            stationary = 'input'
        else:
            raise ValueError("stationary must be 'left' or 'right'")

        if par_left_col is None:
            par_left_col = par_ich

        if par_out_col is None:
            par_out_col = par_och

        if par_left_row is None:
            par_left_row = par_col

        conv2d.conv2d.attribute(self, cshamt_mul, cshamt_sum, cshamt_out,
                                par_left_col, par_out_col, par_left_row, None,
                                concur_out_col, stationary,
                                left_ram_size, right_ram_size,
                                bias_ram_size, scale_ram_size,
                                vshamt_mul_ram_size, vshamt_sum_ram_size, vshamt_out_ram_size,
                                out_ram_size,
                                disable_keep_left)

    def eval(self, memo, input_dict, **kwargs):
        if id(self) in memo:
            return memo[id(self)]

        import nngen.verify as verify

        name = self.__class__.__name__

        args = [arg.eval(memo, input_dict)
                for arg in self.args]

        input = args[0]
        filter = args[1]
        strides = self.strides

        bias = args[self.args_dict['bias']] if self.has_bias else None
        scale = args[self.args_dict['scale']] if self.has_scale else None

        rshift_mul = (args[self.args_dict['vshamt_mul']]
                      if self.has_vshamt_mul else self.cshamt_mul)
        rshift_sum = (args[self.args_dict['vshamt_sum']]
                      if self.has_vshamt_sum else self.cshamt_sum)
        rshift_out = (args[self.args_dict['vshamt_out']]
                      if self.has_vshamt_out else self.cshamt_out)

        kwargs['bias'] = bias
        kwargs['scale'] = scale
        kwargs['transposed_a'] = False
        kwargs['transposed_b'] = True
        kwargs['rshift_mul'] = rshift_mul
        kwargs['rshift_sum'] = rshift_sum
        kwargs['rshift_out'] = rshift_out
        kwargs['act_func'] = self.act_func
        kwargs['asymmetric_clip'] = self.asymmetric_clip
        kwargs['dtype'] = self.dtype
        kwargs['mul_dtype'] = self.mul_dtype
        kwargs['sum_dtype'] = self.sum_dtype
        kwargs['name'] = self.name
        kwargs['par_left_col'] = self.par_left_col
        kwargs['par_left_row'] = self.par_left_row
        kwargs['par_out_col'] = self.par_out_col
        kwargs['concur_out_col'] = self.concur_out_col
        kwargs['stationary'] = self.stationary
        kwargs['a_dtype'] = self.args[0].dtype
        kwargs['b_dtype'] = self.args[1].dtype
        kwargs['bias_dtype'] = self.args[self.args_dict['bias']].dtype if self.has_bias else None
        kwargs['scale_dtype'] = self.args[self.args_dict['scale']].dtype if self.has_scale else None

        method = self.get_eval_method()
        ret = method(input, filter, **kwargs)
        memo[id(self)] = ret

        return ret



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import collections
import numpy as np

import nngen.operator as operator

from . import util


def Squeeze(visitor, node):
    args = []

    for arg in node.input:
        arg_obj = visitor.visit(arg)
        args.append(arg_obj)

    args = [util.optimize_to_raw_value(arg) for arg in args]

    input = args[0]

    if isinstance(input, operator.matmul) and input.onnx_batchnorm is not None:
        return input

    axes = []
    for attribute in node.attribute:
        if attribute.name == 'axes':
            axes = [i for i in attribute.ints]

    if len(axes) == 0:
        for i, s in enumerate(input.shape):
            if i == 1:
                axes.append(i)

    if isinstance(input, (tuple, list)):
        input = np.array(input)

    if isinstance(input, np.ndarray):
        ret = input
        offset = 0

        for axis in sorted(axes):
            ret = np.squeeze(ret, axis - offset)
            offset += 1

        return ret

    layout = input.get_layout()
    onnx_layout = input.get_onnx_layout()

    if layout is not None and onnx_layout is not None:
        axes = [layout.index(onnx_layout[axis]) for axis in axes]
        axes.sort()

    name = util.get_name(node)

    new_shape = input.shape[:]
    offset = 0

    for axis in sorted(axes):
        new_shape.pop(axis - offset)
        offset += 1

    return operator.reshape(input, new_shape, name=name)


def Unsqueeze(visitor, node):
    args = []

    for arg in node.input:
        arg_obj = visitor.visit(arg)
        args.append(arg_obj)

    args = [util.optimize_to_raw_value(arg) for arg in args]

    input = args[0]

    for attribute in node.attribute:
        if attribute.name == 'axes':
            axes = attribute.ints

    name = util.get_name(node)

    ret = input
    offset = 0

    for axis in sorted(axes):
        if isinstance(ret, (tuple, list)):
            ret = np.array(ret)

        if isinstance(ret, (np.ndarray, np.float, np.int, float, int)):
            ret = np.expand_dims(ret, axis + offset)

        else:
            kwargs = collections.OrderedDict()
            kwargs['name'] = '{}_{}'.format(name, offset)
            ret = operator.expand_dims(ret, axis + offset, **kwargs)

        offset += 1

    return ret



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division


def Identity(visitor, node):

    srcs = []

    for src in node.input:
        src_obj = visitor.visit(src)
        srcs.append(src_obj)

    return srcs[0]



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import functools
import math
import numpy as np
from collections import OrderedDict

import nngen.basic_types as bt
from nngen.quantizer import util


class exp(bt._ElementwiseOperator):

    def __init__(self, features,
                 lut_addrwidth=8, lut_clip=6.0, lut_bias=0.0, range_rate=0.95,
                 dtype=None, name=None, par=1):

        shape = None
        if features.dtype is not None and features.dtype.width < 8:
            lut_addrwidth = features.dtype.width

        self.lut_addrwidth = lut_addrwidth
        self.lut_clip = lut_clip
        self.lut_bias = lut_bias
        self.range_rate = range_rate
        bt._ElementwiseOperator.__init__(self, features,
                                         dtype=dtype, shape=shape, name=name, par=par)

    def _get_expected_scale_factor(self):
        return (2 ** (self.lut_addrwidth - 1)) / self.lut_clip

    def _get_features_scale_shamt_bias(self):
        expected_scale_factor = self._get_expected_scale_factor()

        features_scale = np.array([expected_scale_factor / self.args[0].scale_factor])
        q_features_scale, scale_factor = util.quantize_linear_scale(features_scale, 32)
        q_features_scale = int(q_features_scale[0])
        q_features_shamt = round(math.log(scale_factor, 2))
        q_features_bias = round(q_features_scale * self.lut_bias * -1.0)
        return q_features_scale, q_features_shamt, q_features_bias

    def get_local_control_param_values(self):
        (q_features_scale, q_features_shamt,
         q_features_bias) = self._get_features_scale_shamt_bias()
        return OrderedDict([('features_scale_cparam', q_features_scale),
                            ('features_shamt_cparam', q_features_shamt),
                            ('features_bias_cparam', q_features_bias)])

    def get_stream_hash(self):
        base = bt._ActFuncOperator.get_stream_hash(self)
        return (base, self.lut_addrwidth, self.lut_clip, self.lut_bias, self.range_rate)

    def op(self, strm, *args, **kwargs):
        features_signed = self.args[0].get_signed()

        features_scale = strm.ReinterpretCast(self.features_scale_cparam,
                                              width=self.features_scale_cparam.width + 1,
                                              signed=features_signed)
        features_bias = strm.ReinterpretCast(self.features_bias_cparam,
                                             width=self.features_bias_cparam.width + 1,
                                             signed=True)

        mul = strm.Madd(args[0], features_scale, features_bias)
        mul.width = max(mul.width + features_scale.width, features_bias.width)

        features_shamt = strm.ReinterpretCast(self.features_shamt_cparam,
                                              width=self.features_shamt_cparam.width,
                                              signed=False)
        sra = strm.Sra(mul, features_shamt)
        lut_addr = strm.Slice(sra, self.lut_addrwidth - 1, 0)

        addr_scale = 1 / self._get_expected_scale_factor()

        out_width = self.dtype.width
        out_point = self.dtype.point
        out_signed = self.dtype.signed
        if out_signed:
            out_scale = round((2 ** (out_width - 1)) * self.range_rate /
                              np.exp(2 ** (self.lut_addrwidth - 1) * addr_scale + self.lut_bias))
        else:
            out_scale = round((2 ** out_width) * self.range_rate /
                              np.exp(2 ** (self.lut_addrwidth - 1) * addr_scale + self.lut_bias))

        def _exp(x):
            return int(np.around(np.exp(x) * out_scale).astype(np.int64))

        patterns_p = [_exp(i * addr_scale + self.lut_bias)
                      for i in range(2 ** (self.lut_addrwidth - 1))]
        patterns_n = [_exp((-i - 1) * addr_scale + self.lut_bias)
                      for i in range(2 ** (self.lut_addrwidth - 1))]
        patterns_n.reverse()

        patterns = patterns_p + patterns_n

        lut = strm.LUT(lut_addr, patterns, out_width, out_point, out_signed)

        p_th = 2 ** (self.lut_addrwidth - 1) - 1
        n_th = -1 * p_th

        p = strm.Mux(sra > p_th, _exp((p_th + 1) * addr_scale + self.lut_bias), lut)
        n = strm.Mux(sra < n_th, _exp((n_th - 1) * addr_scale + self.lut_bias), lut)
        out = strm.Mux(sra >= 0, p, n)

        return out

    def get_eval_method(self):
        import nngen.verify as verify

        name = self.__class__.__name__
        method = getattr(verify, name, None)

        features_scale, features_shamt, features_bias = self._get_features_scale_shamt_bias()

        method = functools.partial(method,
                                   lut_addrwidth=self.lut_addrwidth,
                                   lut_clip=self.lut_clip,
                                   lut_bias=self.lut_bias,
                                   range_rate=self.range_rate,
                                   features_dtype=self.args[0].dtype,
                                   features_scale=features_scale,
                                   features_shamt=features_shamt,
                                   features_bias=features_bias)
        return method



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from collections import defaultdict
import inspect

from . import basic_types as bt


class _Scheduler(object):
    """ ASAP scheduler """

    def __init__(self):
        self.current_stage = 0
        self.result = defaultdict(list)

    def schedule(self, objs):
        not_scheduled = set()
        not_scheduled.update(objs)

        while not_scheduled:
            candidates = []
            for obj in sorted(not_scheduled, key=lambda x: x.object_id, reverse=True):
                if (self.is_schedulable(obj, self.current_stage) and
                    self.check_constraints(obj) and
                    (obj.parallel_scheduling_allowed or
                     len(self.result[self.current_stage]) == 0)):
                    candidates.append(obj)

            if not candidates:
                self.next_stage()
                continue

            obj = self.select(*candidates)

            obj.set_stage(self.current_stage)
            not_scheduled.remove(obj)

            self.add_result(self.current_stage, obj)

            if not obj.parallel_scheduling_allowed:
                self.next_stage()
                continue

        return self.result

    def is_schedulable(self, obj, stage):
        return obj.is_schedulable(stage)

    def check_constraints(self, obj):
        methods = [method for name, method in inspect.getmembers(self)
                   if inspect.ismethod(method) and
                   name.startswith('constraint_')]

        for method in methods:
            ret = method(obj)
            if not ret:
                return False

        return True

    def next_stage(self):
        self.current_stage += 1

    def select(self, *objs):
        if not objs:
            raise ValueError('no candidate obj')

        return objs[0]

    def add_result(self, stage, obj):
        self.result[stage].append(obj)


class _ListScheduler(_Scheduler):
    """ List scheduler """

    def __init__(self, config):
        _Scheduler.__init__(self)
        self.config = config

    def select(self, *objs):
        if not objs:
            raise ValueError('no candidate')

        return self.select_highest_priority(*objs)

    def select_highest_priority(self, *objs):
        ret = objs[0]
        maxval = self.get_priority(ret)

        for obj in objs[1:]:
            p = self.get_priority(obj)
            if p > maxval:
                ret = obj
                maxval = p

        return ret

    def get_priority(self, obj):
        if bt.is_reduction_operator(obj):
            srcs = obj.collect_sources()
            return len(srcs) + 1

        if bt.is_elementwise_operator(obj):
            srcs = obj.collect_sources()
            return len(srcs) + 2

        return 0


class OperationScheduler(_ListScheduler):
    """ List scheduler """

    def constraint_max_parallel_ops(self, obj):
        if bt.is_storage(obj):
            return True

        if bt.is_output_chainable_operator(obj) and not obj.chain_head:
            return True

        max_parallel_ops = None
        if 'max_parallel_ops' in self.config:
            max_parallel_ops = self.config['max_parallel_ops']

        if max_parallel_ops is None:
            return True

        num_effective_ops = 0
        for op in self.result[self.current_stage]:
            if bt.is_output_chainable_operator(op) and not op.chain_head:
                continue
            num_effective_ops += 1

        if num_effective_ops >= max_parallel_ops:
            return False

        return True



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from . import basic
from . import concat


def normalize(x, y, z, shamt,
              dtype=None, sum_dtype=None, name=None, par=1,
              x_dtype=None, y_dtype=None, z_dtype=None, shamt_dtype=None):

    return basic.multiply_add_rshift_clip(x, y, z, shamt,
                                          dtype, sum_dtype, name, par,
                                          x_dtype, y_dtype, z_dtype, shamt_dtype)


def scaled_add(a, b, a_scale, b_scale, shamt,
               asymmetric_clip=False,
               dtype=None, sum_dtype=None, name=None, par=1,
               a_dtype=None, b_dtype=None):

    v0 = basic.multiply(a, a_scale, dtype=sum_dtype, par=par,
                        x_dtype=a_dtype)
    v1 = basic.multiply(b, b_scale, dtype=sum_dtype, par=par,
                        x_dtype=b_dtype)
    v = basic.add(v0, v1, dtype=sum_dtype, par=par,
                  x_dtype=sum_dtype, y_dtype=sum_dtype)
    v = basic.rshift(v, shamt, dtype=sum_dtype, par=par,
                     x_dtype=sum_dtype)
    return basic.clip(v, asymmetric_clip=asymmetric_clip,
                      dtype=dtype, par=par,
                      x_dtype=sum_dtype)


def scaled_concat(values, scales, shamt, axis,
                  asymmetric_clip=False,
                  dtype=None, mul_dtype=None, name=None):

    scaled_values = []
    for value, scale in zip(values, scales):
        v = basic.multiply(value, scale, dtype=mul_dtype)
        v = basic.rshift(v, shamt, dtype=mul_dtype,
                         x_dtype=mul_dtype)
        v = basic.clip(v, asymmetric_clip=asymmetric_clip,
                       dtype=dtype,
                       x_dtype=mul_dtype)
        scaled_values.append(v)

    return concat(scaled_values, axis, dtype, name)


def scaled_multiply(a, b, shamt,
                    asymmetric_clip=False,
                    dtype=None, mul_dtype=None, name=None, par=1,
                    a_dtype=None, b_dtype=None):

    v = basic.multiply(a, b, dtype=mul_dtype, par=par,
                       x_dtype=a_dtype, y_dtype=b_dtype)
    v = basic.rshift(v, shamt, dtype=mul_dtype, par=par,
                     x_dtype=mul_dtype)
    return basic.clip(v, asymmetric_clip=asymmetric_clip,
                      dtype=dtype, par=par,
                      x_dtype=mul_dtype)


def scaled_div(a, b, shamt,
               asymmetric_clip=False,
               dtype=None, div_dtype=None, name=None, par=1,
               a_dtype=None, b_dtype=None):

    a = basic.lshift(a, shamt, dtype=div_dtype, par=par,
                     x_dtype=a_dtype)
    v = basic.div(a, b, dtype=div_dtype, par=par,
                  x_dtype=a_dtype, y_dtype=b_dtype)
    return basic.clip(v, asymmetric_clip=asymmetric_clip,
                      dtype=dtype, par=par,
                      x_dtype=div_dtype)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import collections
import numpy as np

import nngen.storage as storage
import nngen.operator as operator

from . import util


def _reduce(method, visitor, node, np_method=None):

    args = []

    for src in node.input:
        src_obj = visitor.visit(src)
        args.append(src_obj)

    input = args[0]
    if isinstance(input, (tuple, list)):
        input = np.array(input)

    shape = input.shape

    keepdims = False
    axes = None

    for attribute in node.attribute:
        if attribute.name == 'axes':
            axes = [v for v in attribute.ints]

        elif attribute.name == 'keepdims':
            keepdims = attribute.i != 0

    if axes == [i for i in range(len(shape))]:
        axes = None

    opt_args = [util.optimize_to_raw_value(arg) for arg in args]

    all_ndarray = True
    for arg in opt_args:
        if not isinstance(arg, np.ndarray):
            all_ndarray = False
            break

    if all_ndarray and np_method is not None:
        return np_method(*opt_args, axis=axes, keepdims=keepdims)

    layout = input.get_layout()
    onnx_layout = input.get_onnx_layout()

    if layout is not None and onnx_layout is not None and axes is not None:
        axes = [layout.index(onnx_layout[axis]) for axis in axes]
        axes.sort()

    kwargs = collections.OrderedDict()

    name = util.get_name(node)
    kwargs['name'] = name

    kwargs['axis'] = axes
    kwargs['keep_dims'] = keepdims

    if name in visitor.value_dtypes:
        kwargs['dtype'] = visitor.value_dtypes[name]

    return method(*args, **kwargs)


def _arg(method, visitor, node, np_method=None):

    args = []

    for src in node.input:
        src_obj = visitor.visit(src)
        args.append(src_obj)

    input = args[0]
    if isinstance(input, (tuple, list)):
        input = np.array(input)

    keepdims = False
    axis = 0
    select_last_index = False

    for attribute in node.attribute:
        if attribute.name == 'axis':
            axis = attribute.i

        elif attribute.name == 'keepdims':
            keepdims = attribute.i != 0

        elif attribute.name == 'select_last_index':
            select_last_index = attribute.i != 0

    opt_args = [util.optimize_to_raw_value(arg) for arg in args]

    all_ndarray = True
    for arg in opt_args:
        if not isinstance(arg, np.ndarray):
            all_ndarray = False
            break

    if all_ndarray and np_method is not None:
        return np_method(*opt_args, axis=axes, keepdims=keepdims)

    if select_last_index:
        raise ValueError('select_last_index is not supported.')

    layout = input.get_layout()
    onnx_layout = input.get_onnx_layout()

    if layout is not None and onnx_layout is not None:
        axis = layout.index(onnx_layout[axis])

    kwargs = collections.OrderedDict()

    name = util.get_name(node)
    kwargs['name'] = name

    kwargs['axis'] = axis
    kwargs['keep_dims'] = keepdims

    if name in visitor.value_dtypes:
        kwargs['dtype'] = visitor.value_dtypes[name]

    return method(*args, **kwargs)


def ReduceSum(visitor, node):

    return _reduce(operator.reduce_sum, visitor, node, np.sum)


def ReduceMax(visitor, node):

    return _reduce(operator.reduce_max, visitor, node, np.max)


def ReduceMin(visitor, node):

    return _reduce(operator.reduce_min, visitor, node, np.min)


def ArgMax(visitor, node):

    return _arg(operator.argmax, visitor, node, np.argmax)


def ArgMin(visitor, node):

    return _arg(operator.argmin, visitor, node, np.argmin)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np


def relu(features, dtype=None, name=None, par=1,
         features_dtype=None):

    features_point = 0 if features_dtype is None else features_dtype.point
    out_point = 0 if dtype is None else dtype.point
    out_shift = out_point - features_point

    zeros = np.zeros_like(features, dtype=np.int64)
    comp = features >= 0

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    ret = out_op(np.where(comp, features, zeros))

    return ret


def relu6(features, dtype=None, name=None, par=1,
          features_dtype=None, features_scale_factor=None):

    if features_scale_factor is None:
        raise ValueError('relu6 requires features_scale_factor to adjust the clipping range.')

    features_point = 0 if features_dtype is None else features_dtype.point
    out_point = 0 if dtype is None else dtype.point
    out_shift = out_point - features_point

    zeros = np.zeros_like(features, dtype=np.int64)
    comp0 = features >= 0
    sixs = np.zeros_like(features, dtype=np.int64) + [round(features_scale_factor) * 6]
    comp6 = features > round(features_scale_factor * 6)

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    ret = out_op(np.where(comp0, np.where(comp6, sixs, features), zeros))

    return ret



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from nngen.operator.pool import _pool


class pad(_pool):

    def __sub_str__(self):
        if hasattr(self, 'pad_col_left') and self.padding == 'SAME':
            padding = ("'%s'-(%d, %d, %d, %d)" %
                       (self.padding,
                        self.pad_row_top_value, self.pad_row_bottom_value,
                        self.pad_col_left_value, self.pad_col_right_value))
        else:
            padding = self.padding

        par = ' par:%d' % self.par if self.par > 1 else ''

        value_ram_size = (' value_ram_size:%d' % self.value_ram_size
                          if self.value_ram_size is not None else '')
        out_ram_size = (' out_ram_size:%d' % self.out_ram_size
                        if self.out_ram_size is not None else '')

        return (' padding:%s%s%s%s' %
                (padding, par, value_ram_size, out_ram_size))

    def __init__(self, value, padding,
                 dtype=None, name=None, par=1,
                 value_ram_size=None, out_ram_size=None):

        ksize = (1, 1, 1, 1)
        strides = (1, 1, 1, 1)
        _pool.__init__(self, value, ksize, strides,
                       padding, dtype, name, par,
                       value_ram_size, out_ram_size)

    def get_pad_value(self, strm):
        return strm.Int(0)

    def pool_op(self, strm, index, *vars):
        return vars[0]

    def eval(self, memo, input_dict, **kwargs):
        if id(self) in memo:
            return memo[id(self)]

        import nngen.verify as verify

        name = self.__class__.__name__

        args = [arg.eval(memo, input_dict)
                for arg in self.args]

        value = args[0]

        kwargs['padding'] = self.padding
        kwargs['dtype'] = self.dtype
        kwargs['name'] = self.name
        kwargs['par'] = self.par

        method = self.get_eval_method()
        ret = method(value, **kwargs)
        memo[id(self)] = ret

        return ret
